import os
import kaa
from kaa import doc_re
from kaa import document
from kaa.keyboard import *
from kaa.theme import Style
from kaa.filetype.default import defaultmode
from kaa.command import commandid
from kaa.command import norec
from kaa.command import norerun

FilenameIndexThemes = {
    'basic': [
        Style('filenameindex-filename', 'Green', 'Default'),
        Style('filenameindex-lineno', 'Green', 'Default'),
    ],
}

filenameindex_keys = {
    '\r': ('filenameindex.showmatch'),
    '\n': ('filenameindex.showmatch'),
}


def _enc_japanese(filename):
    encoding = kaa.app.storage.guess_japanese_encoding(filename)
    if encoding:
        return encoding
    else:
        kaa.app.messagebar.set_message(
            'Cannot detect text encoding:: {}'.format(filename))


class FilenameIndexMode(defaultmode.DefaultMode):
    # todo: rename key/command names
    DOCUMENT_MODE = False
    USE_UNDO = False
    HIGHLIGHT_CURSOR_ROW = True

    FILENAMEINDEX_KEY_BINDS = [
        filenameindex_keys,
    ]

    encoding = None
    newline = None

    def init_themes(self):
        super().init_themes()
        self.themes.append(FilenameIndexThemes)

    def init_keybind(self):
        super().init_keybind()
        self.register_keys(self.keybind, self.FILENAMEINDEX_KEY_BINDS)

    def _locate_doc(self, wnd, doc, lineno):
        wnd.show_doc(doc)

        pos = doc.get_lineno_pos(lineno)
        tol = doc.gettol(pos)
        wnd.cursor.setpos(wnd.cursor.adjust_nextpos(wnd.cursor.pos, tol))
        wnd.activate()

    @commandid('filenameindex.showmatch')
    @norec
    @norerun
    def show_hit(self, wnd):
        pos = wnd.cursor.pos
        tol = self.document.gettol(pos)
        eol, line = self.document.getline(tol)

        try:
            filename, lineno, line = line.split(':', 2)
        except ValueError:
            return

        if not filename:
            return

        try:
            lineno = int(lineno)
        except ValueError:
            lineno = 1

        filename = os.path.abspath(os.path.expanduser(filename))
        doc = document.Document.find_filename(filename)
        if not doc:
            enc = self.encoding
            if not enc:
                enc = kaa.app.config.DEFAULT_ENCODING
            if enc == 'japanese':
                enc = _enc_japanese(filename)

            newline = self.newline
            if not newline:
                newline = kaa.app.config.DEFAULT_NEWLINE

            doc = kaa.app.storage.openfile(
                filename, encoding=enc, newline=newline)

        buddy = wnd.splitter.get_buddy()
        if not buddy:
            buddy = wnd.splitter.split(vert=False, doc=doc)
            self._locate_doc(buddy.wnd, doc, lineno)
        else:
            if buddy.wnd and buddy.wnd.document is doc:
                self._locate_doc(buddy.wnd, doc, lineno)
                return

            def callback(canceled):
                if not canceled:
                    buddy.show_doc(doc)
                    self._locate_doc(buddy.wnd, doc, lineno)
            kaa.app.app_commands.save_splitterdocs(wnd, buddy, callback)

    RE_FILENAME = doc_re.compile(
        r'(?P<FILENAME>^[^:\n]+)\:(?P<LINENO>\d+)\:.*$',
        doc_re.M | doc_re.X)

    def is_match(self, pos):
        m = self.RE_FILENAME.match(self.document, pos)
        return m

    def on_global_prev(self, wnd):
        if kaa.app.focus in self.document.wnds:
            if self.is_match(self.document.gettol(wnd.cursor.pos)):
                self.show_hit(kaa.app.focus)
                return True

        pos = wnd.cursor.pos

        while True:
            eol = self.document.gettol(pos)
            if eol:
                tol = self.document.gettol(eol - 1)
            else:
                eol = self.document.endpos()
                tol = self.document.gettol(eol)

            if self.is_match(tol):
                wnd.cursor.setpos(tol)
                self.show_hit(wnd)
                return True

            if tol == 0:
                wnd.cursor.setpos(tol)
                self.document.wnds[0].activate()
                return True

            pos = tol

    def on_global_next(self, wnd):
        if kaa.app.focus in self.document.wnds:
            if self.is_match(self.document.gettol(wnd.cursor.pos)):
                self.show_hit(kaa.app.focus)
                return True

        pos = wnd.cursor.pos
        while True:
            tol = self.document.geteol(pos)
            m = self.is_match(tol)
            if m:
                wnd.cursor.setpos(tol)
                self.show_hit(wnd)
                return True

            if self.document.geteol(tol) == self.document.endpos():
                wnd.cursor.setpos(0)
                self.document.wnds[0].activate()
                return True

            pos = tol



import copy
import time
from collections import defaultdict, namedtuple
import kaa
from . import keybind, theme, modebase, menu


class DefaultMode(modebase.ModeBase):
    DOCUMENT_MODE = True
    MODENAME = 'default'
    SHOW_LINENO = False
    SHOW_BLANK_LINE = True
    VI_COMMAND_MODE = False

    KEY_BINDS = [
        keybind.app_keys,
        keybind.cursor_keys,
        keybind.edit_command_keys,
        keybind.addtional_edit_command_keys,
        keybind.emacs_keys,
        keybind.search_command_keys,
        keybind.macro_command_keys,
        keybind.rerun_keys,
    ]

    VI_KEY_BIND = [
        keybind.command_mode_keys
    ]

    VI_VISUAL_MODE_KEY_BIND = [
        keybind.visual_mode_keys
    ]
    VI_VISUAL_LINEWISE_MODE_KEY_BIND = [
        keybind.visual_linewise_mode_keys
    ]

    def init_keybind(self):
        super().init_keybind()
        self.register_keys(self.keybind, self.KEY_BINDS)
        self.register_keys(self.keybind_vi_commandmode, self.VI_KEY_BIND)
        self.register_keys(self.keybind_vi_visualmode,
                           self.VI_VISUAL_MODE_KEY_BIND)
        self.register_keys(self.keybind_vi_visuallinewisemode,
                           self.VI_VISUAL_LINEWISE_MODE_KEY_BIND)

    def init_menu(self):
        self.menu = copy.deepcopy(menu.MENUS)

    def init_themes(self):
        super().init_themes()
        self.themes.append(theme.DefaultThemes)

    def close(self):
        super().close()
        self.keybind_vi_commandmode.clear()
        self.keybind_vi_visualmode.clear()
        self.keybind_vi_visuallinewisemode.clear()

    def on_idle(self):
        if self.closed:
            return

        ret = super().on_idle()
        if not ret:
            ret = self.check_fileupdate()

        return ret

    INTERVAL_CHECKUPDATE = 10

    def check_fileupdate(self):
        if not self.DOCUMENT_MODE:
            return
        if not kaa.app.mainframe.is_idle():
            return

        t = time.time()
        if t - self._check_fileupdate < self.INTERVAL_CHECKUPDATE:
            return

        self._check_fileupdate = t
        if self.document.fileinfo:
            if self.document.fileinfo.check_update():
                kaa.app.file_commands.notify_fileupdated(self.document)

    def on_esc_pressed(self, wnd, event):
        # Pressing esc key starts command mode.
        if self.VI_COMMAND_MODE:
            is_available, command = self.get_command('editmode.command')
            if command:
                command(wnd)
                if kaa.app.macro.is_recording():
                    kaa.app.macro.record(1, command)

    def on_keypressed(self, wnd, event, s, commands, candidate):
        if not commands and not candidate:
            if not s or s[0] < ' ':
                msg = self.DEFAULT_MENU_MESSAGE or kaa.app.DEFAULT_MENU_MESSAGE
                kaa.app.messagebar.set_message(msg)

        return super().on_keypressed(wnd, event, s, commands, candidate)

    def _show_parenthesis(self, charattrs, pos):
        charattrs[pos] = self.get_styleid('parenthesis_cur')
        matchpos = self.find_match_parenthesis(pos)
        if matchpos is not None:
            charattrs[matchpos] = self.get_styleid('parenthesis_match')

    def update_charattr(self, wnd):
        pos = wnd.cursor.pos
        d = {}
        c = ''
        if pos < self.document.endpos():
            c = self.document.buf[pos]

        if c and (c in self.PARENTHESIS):
            self._show_parenthesis(d, pos)
        elif 1 < pos:
            c = self.document.buf[pos - 1]
            if c in self.PARENTHESIS_CLOSE:
                self._show_parenthesis(d, pos - 1)

        if d != wnd.charattrs:
            wnd.charattrs = d
            wnd.screen.style_updated()
            return True

    PARENTHESIS_OPEN = '({['
    PARENTHESIS_CLOSE = ')}]'
    PARENTHESIS = PARENTHESIS_OPEN + PARENTHESIS_CLOSE
    PARENSIS_PAIR = {o: c for (o, c) in
                     zip(PARENTHESIS_OPEN + PARENTHESIS_CLOSE,
                         PARENTHESIS_CLOSE + PARENTHESIS_OPEN)}

    def iter_parenthesis(self, posfrom):
        while True:
            pos = self.document.buf.findchr(
                self.PARENTHESIS, posfrom, self.document.endpos())

            if pos == -1:
                break

            attr = self.document.styles.getint(pos)
            yield pos, self.document.buf[pos], attr
            posfrom = pos + 1

    def iter_rev_parenthesis(self, posfrom):
        posfrom += 1
        while True:
            pos = self.document.buf.rfindchr(
                self.PARENTHESIS, 0, posfrom)

            if pos == -1:
                break

            attr = self.document.styles.getint(pos)
            yield pos, self.document.buf[pos], attr
            posfrom = pos

    def find_match_parenthesis(self, posfrom):
        opener = self.document.buf[posfrom]
        curattr = self.document.styles.getint(posfrom)

        d = defaultdict(int)
        if opener in self.PARENTHESIS_OPEN:
            f = self.iter_parenthesis
            key = (opener, curattr)
        else:
            f = self.iter_rev_parenthesis
            key = (self.PARENSIS_PAIR[opener], curattr)

        for pos, c, attr in f(posfrom):
            if c in self.PARENTHESIS_OPEN:
                d[(c, attr)] += 1
            else:
                d[(self.PARENSIS_PAIR[c], attr)] -= 1

            if d.get(key) == 0:
                return pos

    _headerinfo = namedtuple('_headerinfo',
                             ['token', 'parent', 'name', 'dispname', 'lineno', 'pos'])

    class HeaderInfo(_headerinfo):

        def get_parents(self):
            ret = []
            p = self.parent
            while p:
                ret.insert(0, p)
                p = p.parent
            return ret

    def get_headers(self):
        return ()



import unicodedata
import re
import kaa
from kaa import document
from kaa.ui.dialog import dialogmode
from kaa.theme import Style

MsgBoxThemes = {
    'basic': [
        Style('underline', 'Base3', 'Base02', underline=True),
        Style('separator', 'LightBlue', 'Base02', nowrap=False),
        Style('button', 'LightBlue', 'Base02', nowrap=True),
        Style('button.shortcut', 'LightBlue', 'Base02', underline=True,
              nowrap=True),
    ],
}


class MsgBoxMode(dialogmode.DialogMode):
    autoshrink = True
    SEPARATOR = '/'
    USE_UNDO = False
    DELAY_STR = False

    def init_themes(self):
        super().init_themes()
        self.themes.append(MsgBoxThemes)

    def init_keybind(self):
        pass

    def on_add_window(self, wnd):
        super().on_add_window(wnd)

        wnd.CURSOR_TO_MIDDLE_ON_SCROLL = False

    def on_dialog_start(self, wnd):
        wnd.cursor.setpos(self.document.endpos() - 1)

    def on_str(self, wnd, s, overwrite=False):
        s = unicodedata.normalize('NFKC', s).lower()
        for c in s:
            if c in self.shortcuts:
                self.on_shortcut(wnd, c)
                return
            if self.keys and (c in self.keys):
                self.on_shortcut(wnd, c)
                return

    def on_shortcut(self, wnd, c):
        # return value
        if c:
            c = c.lower()
        try:
            self._runcallback(c)
        finally:
            # Destroy popup window
            if wnd:
                popup = wnd.get_label('popup')
                if popup:
                    popup.destroy()

    def _runcallback(self, c):
        if self.callback:
            self.callback(c)

    def on_esc_pressed(self, wnd, event):
        self.on_shortcut(wnd, None)

    def _show_window(self):
        kaa.app.show_dialog(self.document)

    @classmethod
    def build_msgbox(cls, caption, options, callback, keys=None, border=False):
        doc = document.Document()
        mode = cls()
        mode.callback = callback
        mode.keys = keys
        mode.border = border
        doc.setmode(mode)

        with dialogmode.FormBuilder(doc) as f:

            # caption
            if caption:
                f.append_text('caption', caption)
                f.append_text('default', ' ')

            mode.shortcuts = {}
            for n, option in enumerate(options):
                m = re.search(r'&([^&])', option)
                shortcut = m.group(1)
                mode.shortcuts[shortcut.lower()] = option

                f.append_text('button',
                              option,
                              on_shortcut=lambda wnd, key=shortcut: mode.on_shortcut(
                                  wnd, key),
                              shortcut_style='button.shortcut')

                if n < len(options) - 1:
                    f.append_text('separator', cls.SEPARATOR)

            f.append_text('default', ' ')
            f.append_text('underline', ' ')

        return doc

    @classmethod
    def show_msgbox(cls, caption, options, callback, keys=None, border=False):
        doc = cls.build_msgbox(caption, options, callback, keys, border)
        doc.mode._show_window()
        return doc



import kaa
from kaa.ui.selectlist import selectlist, filterlist
from kaa.keyboard import *

texthist_keys = {
    '\r': 'texthist.select',
    '\n': 'texthist.select',
}


class TextHistMode(filterlist.FilterListMode):
    MAX_CAPTION_LEN = 256
    SCREEN_NOWRAP = True

    def init_keybind(self):
        super().init_keybind()
        self.keybind.add_keybind(texthist_keys)

    def set_candidates(self, candidates):
        self.candidates = []
        for c in candidates:
            caption = ' '.join(c.split())
            if not caption:
                caption = '(...)'
            if len(caption) > self.MAX_CAPTION_LEN:
                caption = caption[:self.MAX_CAPTION_LEN] + '...'
            c = selectlist.SelectItem(
                'selectitem', 'selectitem-active', caption, c)
            self.candidates.append(c)


def show_history(title, callback, words):
    doc = filterlist.FilterListInputDlgMode.build(
        title, callback)
    dlg = kaa.app.show_dialog(doc)

    filterlistdoc = TextHistMode.build()
    list = dlg.add_doc('dlg_filterlist', 0, filterlistdoc)

    filterlistdoc.mode.set_candidates(words)
    filterlistdoc.mode.set_query(list, '')
    dlg.on_console_resized()

    return dlg



import kaa
from kaa import document, command
from kaa.ui.dialog import dialogmode
from kaa.filetype.default import keybind
from kaa.keyboard import *

multiline_keys = {
    (alt, '\r'): ('multiline.done'),
    (alt, '\n'): ('multiline.done'),
}


class MultilineMode(dialogmode.DialogMode):
    autoshrink = True
    USE_UNDO = True
    auto_indent = False

    callback = None

    KEY_BINDS = [
        keybind.cursor_keys,
        keybind.edit_command_keys,
        keybind.emacs_keys,
        keybind.macro_command_keys,
        multiline_keys
    ]

    def init_keybind(self):
        super().init_keybind()

        self.register_keys(self.keybind, self.KEY_BINDS)

    def on_add_window(self, wnd):
        super().on_add_window(wnd)

        wnd.set_cursor(dialogmode.DialogCursor(wnd,
                                               [dialogmode.MarkRange('multiline')]))

        wnd.cursor.setpos(self.document.marks['multiline'][0])

    def on_esc_pressed(self, wnd, event):
        popup = wnd.get_label('popup')
        popup.destroy()
        kaa.app.messagebar.set_message("Canceled")

        if self.callback:
            self.callback(None)

    @command.commandid('multiline.done')
    @command.norec
    @command.norerun
    def multiline_done(self, wnd):
        if self.callback:
            f, t = self.document.marks['multiline']
            s = self.document.gettext(f, t)

        popup = wnd.get_label('popup')
        popup.destroy()

        if self.callback:
            self.callback(s)

    CAPTION = 'Paste text here. Hit alt+Enter when finished:'

    @classmethod
    def build(cls, init_text=''):
        doc = document.Document()
        mode = cls()
        doc.setmode(mode)

        with dialogmode.FormBuilder(doc) as f:
            # caption
            f.append_text('caption', cls.CAPTION)
            f.append_text('default', '\n')
            f.append_text('default', init_text, mark_pair='multiline')

        return doc



import locale
from kaa.ui.multiline import multilinemode
from kaa.keyboard import *


class CommitDialogMode(multilinemode.MultilineMode):

    CAPTION = 'Hit alt+Enter to commit'

    commit_callback = None

    def callback(self, s):
        self.conn.send(b'done')
        self.conn.close()
        self.conn = None
        if s is not None:
            open(self.commitmsg, 'w', encoding=locale.getpreferredencoding()).write(s)

        self.commit_callback()



""" Utility functions for kaa customization."""

from kaa.keyboard import *
import kaa.command
import kaa.config


def setup(filemode='kaa.filetype.default.defaultmode.DefaultMode'):
    assert isinstance(filemode, str)

    def _f(f):
        add_class_addon(filemode, 'setup', f)
        return f
    return _f


# Dict to register mode class addon.
# Mode class fetches addon functions form this dict.
_class_addons = {}


def add_class_addon(classname, key, value):
    """Register class addon"""

    d = _class_addons.setdefault(classname, {})
    d.setdefault(key, []).append(value)


def get_addon(name, key, default=()):
    """Get class addon"""
    d = _class_addons.get(name, None)
    if d:
        return d.get(key, default)
    return default


def keybind(filemode='kaa.filetype.default.defaultmode.DefaultMode',
            editmode='input', keymap=None):
    """Register keybind.

    filemode -- Mode class name to register keybind.
                  (default 'kaa.filetype.default.defaultmode.DefaultMode)
    editmode -- Editmode to register keybind.
                  Valid values are:'insert', 'command', 'visual', 'visualline'.
                  (default 'input')

    keymap -- Dictionary of keybind and command name.

    ex)
        keybind(
             filemode='kaa.filetype.python.pythonmode.PythonMode',
             keys = {
                    ((ctrl, 'q'), '2'):
                        'editor.splithorz'})   # Assign C-q 2 to split window.
    """

    add_class_addon(filemode, 'keybind', (editmode, keymap))


def command(commandid,
            filemode='kaa.filetype.default.defaultmode.DefaultMode'):
    """Decorator to register command.

    commandid -- command id
    filemode -- Mode class name to register keybind.
                  (default 'kaa.filetype.default.defaultmode.DefaultMode)

    ex)
        @command(filemode='kaa.filetype.python.pythonmode.PythonMode',
                 commandid='abc.def')
        @norec
        def abc_def(wnd):
            pass
    """

    def _f(f):
        kaa.command.commandid(commandid)(f)
        add_class_addon(filemode, 'command', f)
        return f
    return _f


# theme_def(filetype='kaa.filetype.python', theme={
#   'basic': [
#    	Sytle('default', 'white', 'black'),
#   ])

def theme_def(filemode='kaa.filetype.default.defaultmode.DefaultMode',
              theme=None):
    """Update theme color.

    filemode -- Mode class name to register keybind.
                  (default 'kaa.filetype.default.defaultmode.DefaultMode)
    theme -- Dictionary of theme name and list of styles. Currently, the only
             valid theme name is 'basic'.

    ex)
        # update default foreground color to red.
        theme_def(theme={
            'basic': [
               Style('default', 'red', None),
            ]})

    """

    add_class_addon(filemode, 'style', theme)


# add_filetype('my.python.filetype')

def add_filetype(name):
    """Register filetype package.

    name -- name of filetype package in string.
    """

    kaa.config.FILETYPES.append(name)



import curses
import code
import sys
import time
from contextlib import contextmanager
import kaa
from kaa import document
from kaa.keyboard import *
from kaa.filetype.python import pythonmode
from kaa.command import commandid
from kaa.command import norec
from kaa.command import norerun
from kaa.theme import Style

pythonconsole_keys = {
    ('\r'): 'pythonconsole.newline',
    (alt, '\r'): 'pythonconsole.script-history',
    (alt, '\n'): 'pythonconsole.script-history',
}

PythonConsoleThemes = {
    'basic': [
        Style('stdout', 'Orange', 'Default'),
        Style('stderr', 'Magenta', 'Default'),
        Style('ps', 'Blue', 'Default'),
    ],
}


class KaaInterpreter(code.InteractiveInterpreter):
    localdict = {"__name__": "__console__", "__doc__": None}

    def __init__(self, document):
        super().__init__(self.localdict)
        self.document = document

    def runcode(self, code):
        # stop undo
        self.document.use_undo(False)
        # send newline before running script
        sys.stdout.write('\n')
        try:
            super().runcode(code)
        finally:
            # stop undo
            self.document.use_undo(True)

    def showsyntaxerror(self, filename=None):
        # stop undo
        self.document.use_undo(False)
        # send newline before show error
        sys.stdout.write('\n')
        try:
            super().showsyntaxerror(filename)
        finally:
            # stop undo
            self.document.use_undo(True)


class PythonConsoleMode(pythonmode.PythonMode):
    DEFAULT_STATUS_MSG = 'Hit alt+Enter for history.'
    MODENAME = 'Python console'
    DOCUMENT_MODE = False
    SHOW_BLANK_LINE = False

    PYTHONCONSOLE_KEY_BINDS = [
        pythonconsole_keys,
    ]

    def init_keybind(self):
        super().init_keybind()
        self.register_keys(self.keybind, self.PYTHONCONSOLE_KEY_BINDS)

    def init_themes(self):
        super().init_themes()
        self.themes.append(PythonConsoleThemes)

    def on_set_document(self, document):
        super().on_set_document(document)
        self.interp = KaaInterpreter(document=document)

        self.document.append(
            'Python %s\n' %
            sys.version, self.get_styleid('ps'))
        self.document.append('>>>', self.get_styleid('ps'))
        self.document.append('\n', self.get_styleid('default'))
        p = self.document.endpos()
        self.document.marks['current_script'] = (p, p)

    def on_add_window(self, wnd):
        super().on_add_window(wnd)

        f, t = self.document.marks['current_script']
        wnd.cursor.setpos(f)

    def _get_highlight_range(self):
        return self.document.marks['current_script']

    @commandid('pythonconsole.clear')
    @norec
    @norerun
    def clear_line(self, wnd):
        f, t = self.document.marks['current_script']
        self.document.delete(f, t)
        wnd.cursor.setpos(f)
        self.document.undo.clear()

    def on_esc_pressed(self, wnd, event):
        self.clear_line(wnd)

    def put_string(self, wnd, s, overwrite=False):
        pos = wnd.cursor.pos
        f, t = self.document.marks['current_script']
        if not (f <= pos <= t):
            wnd.cursor.setpos(t)
            wnd.screen.selection.clear()

        super().put_string(wnd, s, overwrite)

    def replace_string(self, wnd, pos, posto, s, update_cursor=True):
        f, t = self.document.marks['current_script']
        if not (f <= pos <= t):
            return
        if not (f <= posto <= t):
            return

        super().replace_string(wnd, pos, posto, s, update_cursor)

    def delete_string(self, wnd, pos, posto, update_cursor=True):
        f, t = self.document.marks['current_script']
        if not (f <= pos <= t):
            return
        if not (f <= posto <= t):
            return

        super().delete_string(wnd, pos, posto, update_cursor)

    @contextmanager
    def _redirect_output(self, wnd):
        stdin = sys.stdin
        stdout = sys.stdout
        stderr = sys.stderr

        style_stdout = self.get_styleid('stdout')
        style_stderr = self.get_styleid('stderr')

        class out:

            def __init__(self, document, style):
                self.document = document
                self.style = style
                self._cache = []

            def write(self, s):
                if self.document.closed:
                    return
                self._cache.append(s)
                if '\n' in s:
                    self.flush()

            def flush(self):
                if self._cache:
                    self.document.append(''.join(self._cache), self.style)
                    self._cache = []

        sys.stdin = None
        sys.stdout = out(self.document, style_stdout)
        sys.stderr = out(self.document, style_stderr)

        try:
            yield

        finally:
            sys.stdout.flush()
            sys.stderr.flush()

            sys.stdin = stdin
            sys.stdout = stdout
            sys.stderr = stderr

    @commandid('pythonconsole.exec')
    @norec
    @norerun
    def exec_script(self, wnd):
        f, t = self.document.marks['current_script']
        s = wnd.document.gettext(f, t).lstrip().rstrip(' \t')
        with self._redirect_output(wnd):
            curses.cbreak()
            try:
                start = time.time()
                ret = self.interp.runsource(s)
                end = time.time()
            finally:
                curses.raw()

            if not ret:
                kaa.app.messagebar.set_message(
                    'Execued in {} secs'.format(end - start))
                wnd.document.undo.clear()
                if s.strip():
                    hist = kaa.app.config.hist(
                        'pyconsole_script',
                        self.MAX_HISTORY)
                    hist.add(s)

                self.document.append('>>>', self.get_styleid('ps'))
                self.document.append('\n', self.get_styleid('default'))
                p = self.document.endpos()
                self.document.marks['current_script'] = (p, p)
                wnd.cursor.setpos(p)
            else:
                self.on_commands(wnd, ['edit.newline'])
                return

    @commandid('pythonconsole.newline')
    def on_enter(self, wnd):
        f, t = self.document.marks['current_script']

        if wnd.cursor.pos < f:
            wnd.cursor.setpos(f)
            return

        tail = wnd.document.gettext(wnd.cursor.pos, t).strip()
        if tail:
            self.on_commands(wnd, ['edit.newline'])
            return

        self.exec_script(wnd)

    def _put_script(self, wnd, text):
        if text.endswith('\n'):
            text = text[:-1]

        self.put_string(wnd, text)
        wnd.screen.selection.clear()
        hist = kaa.app.config.hist('pyconsole_script', self.MAX_HISTORY)
        hist.add(text)

    MAX_HISTORY = 100

    @commandid('pythonconsole.script-history')
    @norec
    @norerun
    def history(self, wnd):
        hist = kaa.app.config.hist('pyconsole_script', self.MAX_HISTORY)
        scripts = [p for p, i in hist.get()]
        if not scripts:
            return

        def callback(text):
            if text:
                self._put_script(wnd, text)

        from kaa.ui.texthist import texthistmode
        texthistmode.show_history('Search history:', callback, scripts)

    @commandid('edit.paste')
    def paste(self, wnd):
        s = kaa.app.clipboard.get()
        if s:
            self._put_script(wnd, s)


def show_console():
    cons = document.Document()
    mode = PythonConsoleMode()
    cons.setmode(mode)
    cons.set_title('<Python console>')

    kaa.app.show_doc(cons)
    kaa.app.messagebar.set_message('')



from pathlib import Path
from git import Repo

def open_repo(d):
    cur = Path(d).absolute()
    while not cur.joinpath('.git').is_dir():
        p = cur.parent
        if p == cur:
            raise RuntimeError('Not a git repogitory')
        cur = p

    repo_dir = str(cur)
    repo = Repo(repo_dir)
    return repo


import kaa
from kaa.command import Commands, commandid


class EditModeCommands(Commands):

    @commandid('editmode.command')
    def editmode_commandmode(self, wnd):
        wnd.document.mode.editmode_command(wnd)
        kaa.app.messagebar.set_message(
            'You are in command mode now. Type `i` to return insert mode.')

    @commandid('editmode.insert')
    def editmode_insertmode(self, wnd):
        wnd.document.mode.editmode_insert(wnd)
        kaa.app.messagebar.set_message(
            'You are in insert mode now. Type `ESC` to return command mode.')

    @commandid('editmode.replace')
    def editmode_replacemode(self, wnd):
        wnd.document.mode.editmode_replace(wnd)
        kaa.app.messagebar.set_message(
            'You are in replace mode now. Type `ESC` to return command mode.')

    @commandid('editmode.visual')
    def editmode_visualmode(self, wnd):
        wnd.screen.selection.clear()
        wnd.document.mode.editmode_visual(wnd)

    @commandid('editmode.visual-linewise')
    def editmode_visualmode_linewise(self, wnd):
        wnd.screen.selection.clear()
        tol = wnd.cursor.adjust_nextpos(
            wnd.cursor.pos, wnd.document.gettol(wnd.cursor.pos))
        wnd.screen.selection.begin_cursor_sel(tol)
        wnd.document.mode.editmode_visual_linewise(wnd)



from unittest.mock import patch

import kaa
from kaa.ui.wordcomplete import wordcompletemode
import kaa_testutils


class TestSearchDlgMode(kaa_testutils._TestDocBase):
    def _run(self, s, pos):
        target = kaa_testutils._TestScreenBase()._getwnd(s)
        doc = wordcompletemode.WordCompleteInputMode.build(target)
        return doc.mode._get_target_word(pos)

    def test_wordcomplete(self):
        assert self._run('hello', 0) == (0, 5)
        assert self._run('hello', 5) == (0, 5)
        assert self._run('', 0) == (0, 0)

    def test_punctuation(self):
        assert self._run('self.', 5) == (5, 5)

    def test_punctuation2(self):
        assert self._run('v[', 1) == (0, 1)




from unittest.mock import patch

import kaa
from kaa.ui.msgbox import msgboxmode
import kaa_testutils


class TestSearchDlgMode(kaa_testutils._TestDocBase):

    def test_msgbox(self):
        kaa.app.DEFAULT_THEME = 'basic'

        def cb(c):
            pass
        doc = msgboxmode.MsgBoxMode.show_msgbox(
            'caption', ['&Yes', '&No', '&All', '&Cancel'], cb)

        assert doc.mode.shortcuts == {
            'y': '&Yes', 'n': '&No', 'a': '&All', 'c': '&Cancel'
        }



import curses
import os
import fnmatch
import re

import kaa
from kaa import document
from kaa.theme import Style
from kaa.ui.filenameindex import filenameindexmode

FILES_IGNORE = [
    '*.com',
    '*.class',
    '*.dll',
    '*.exe',
    '*.o',
    '*.a',
    '*.so',
    '*.pyc',
    '*.pyo',

    '*.7z',
    '*.dmg',
    '*.gz',
    '*.bz2',
    '*.iso',
    '*.jar',
    '*.rar',
    '*.tar',
    '*.zip',

    '*.png',
    '*.gif',
    '*.jpeg',
    '*.jpg',
    '*.bmp',

    '.DS_Store',
    '.Trashes',
    'ehthumbs.db',
    'Thumbs.db',

    '*.bak',
    '#*#',
    '*.swp',
]

DIRS_IGNORE = [
    '.git',
    '.hg',
    '.bzr',
    '.svn',
    'CVS',
    '__pycache__',
]


def _build_fnre(filenames):
    res = [fnmatch.translate(name) for name in filenames.split(' ')]
    return re.compile('|'.join(res))


def _walk(dirname, filenames, tree):
    fn = _build_fnre(filenames)

    ignores = [fnmatch.translate(name) for name in FILES_IGNORE]
    fn_ignore = re.compile('|'.join(ignores))

    ignores = [fnmatch.translate(name) for name in DIRS_IGNORE]
    dir_ignore = re.compile('|'.join(ignores))

    for root, dirs, files in os.walk(dirname, topdown=True):
        files = [os.path.join(root, name) for name in files
                 if fn.match(name) and not fn_ignore.match(name)]
        yield from files
        if not tree:
            return

        dirs[:] = [dir for dir in dirs if not dir_ignore.match(dir)]


def _iter_lines(text):
    lineno = 1
    pos = 0
    while True:
        next = text.find('\n', pos)
        if next == -1:
            break
        next = next + 1
        yield lineno, pos, next, text[pos:next]
        lineno += 1
        pos = next

    yield lineno, pos, len(text), text[pos:]


def _enc_japanese(filename):
    encoding = kaa.app.storage.guess_japanese_encoding(filename)
    if encoding:
        return encoding
    else:
        kaa.app.messagebar.set_message(
            'Cannot detect text encoding:: {}'.format(filename))


def _grep(filename, regex, encoding, newline):
    doc = document.Document.find_filename(filename)
    if doc:
        text = doc.gettext(0, doc.endpos())
    else:
        if encoding == 'japanese':
            encoding = _enc_japanese(filename)

        fileinfo = kaa.app.storage.get_fileinfo(filename, encoding=encoding,
                                                newline=newline)
        f = kaa.app.storage.get_textio(fileinfo)
        if not f:
            return
        with f:
            text = f.read()

    lines = _iter_lines(text)
    lineno, linefrom, lineto, line = next(lines)

    for m in regex.finditer(text):
        f, t = m.span()
        while f >= lineto:
            lineno, linefrom, lineto, line = next(lines)

        yield lineno, linefrom, lineto, line, f, t, text[f:t]


def _search(dir, option, doc):
    mode = doc.mode
    style_default = mode.get_styleid('default')
    style_filename = mode.get_styleid('filenameindex-filename')
    style_lineno = mode.get_styleid('filenameindex-lineno')
    style_match = mode.get_styleid('grep-match')

    def add_hit():
        doc.append(path, style_filename)
        doc.append(':', style_filename)
        doc.append(str(cur_lineno), style_lineno)
        doc.append(':', style_lineno)
        doc.append(' ', style_default)

        linefrom, lineto = linerange
        pos = f = t = 0
        for f, t in matches:
            f -= linefrom
            t -= linefrom
            if f != pos:
                doc.append(cur_line[pos:f], style_default)

            doc.append(cur_line[f:t], style_match)
            pos = t

        if pos != len(cur_line):
            doc.append(cur_line[t:], style_default)

        doc.append('\n', style_default)

    files = _walk(dir, option.filenames, option.tree)
    regex = option.get_regex()

    nfiles = 0
    nhits = 0

    # todo: use raw mode to accept ^C
    for fname in files:
        print(fname)

        nfiles += 1

        cur_lineno = None
        matches = []

        path = kaa.utils.shorten_filename(fname)
        if len(path) > len(fname):
            path = fname

        for rec in _grep(fname, regex, option.encoding, option.newline):
            lineno, linefrom, lineto, line, f, t, match = rec
            nhits += 1

            line = line.rstrip('\n')
            if cur_lineno is None:
                cur_lineno = lineno
                linerange = linefrom, lineto
                cur_line = line

            if matches and (lineno != cur_lineno):
                add_hit()
                matches = []
                linerange = linefrom, lineto
                cur_lineno = lineno
                cur_line = line

            matches.append((f, t))

        if matches:
            add_hit()
    return nfiles, nhits


def grep(option, target):

# todo: reuse grep pane in same frame
#        buddy = wnd.splitter.get_buddy()
#        if not buddy:
#            buddy = wnd.splitter.split(vert=False, doc=doc)
#            self._locate_doc(buddy.wnd, doc, lineno)
#        else:
#            if buddy.wnd and buddy.wnd.document is doc:
#                self._locate_doc(buddy.wnd, doc, lineno)
#                return
#

    if not target:
        doc = document.Document()
        mode = GrepMode()
        doc.setmode(mode)
    else:
        doc = target.document
        doc.delete(0, doc.endpos())
        mode = doc.mode

    doc.set_title('<grep>')
    mode.grepoption = option.clone()
    mode.encoding = mode.grepoption.encoding
    mode.newlilne = mode.grepoption.newline

    dir = os.path.abspath(os.path.expanduser(option.directory))
    if not os.path.isdir(dir):
        s = 'Cannot find directory `{}`.'.format(dir)
        doc.append(s)
    else:
        try:
            # Restore to cooked mode to accept ^C.
            curses.def_prog_mode()
            curses.endwin()

            try:
                print('Hit ^C to cancel grep.')
                nfiles, nhits = _search(dir, option, doc)
            finally:
                curses.reset_prog_mode()
                kaa.app.mainframe.refresh()

        except KeyboardInterrupt:
            kaa.app.messagebar.set_message('Grep canceled')
            return

        if not nfiles:
            s = 'Cannot find file `{}`.'.format(option.filenames)
            doc.append(s)
        elif not nhits:
            s = 'Cannot find `{}`.'.format(option.text)
            doc.append(s)
        else:
            s = 'Found {} times in {} files'.format(nhits, nfiles)

    kaa.app.messagebar.set_message(s)

    if not target:
        kaa.app.show_doc(doc)
    else:
        target.activate()


GrepThemes = {
    'basic': [
        Style('grep-match', 'Orange', None),
    ],
}


class GrepMode(filenameindexmode.FilenameIndexMode):
    MODENAME = 'Grep'

    def init_themes(self):
        super().init_themes()
        self.themes.append(GrepThemes)



from kaa import document
import kaa_testutils


class TestDocument(kaa_testutils._TestScreenBase):

    def test_gettol(self):
        assert self._getdoc('').gettol(0) == 0
        assert self._getdoc('0123').gettol(4) == 0
        assert self._getdoc('0123\n').gettol(4) == 0
        assert self._getdoc('012\n').gettol(4) == 4

    def test_geteol(self):
        assert self._getdoc('').geteol(0) == 0
        assert self._getdoc('0123').geteol(0) == 4
        assert self._getdoc('0123\n').geteol(0) == 5
        assert self._getdoc('012\r3').geteol(0) == 5
        assert self._getdoc('012\r\n').geteol(0) == 5
        assert self._getdoc('012\r\n').geteol(3) == 5
        assert self._getdoc('012\r\n').geteol(4) == 5

    def test_getline(self):
        assert self._getdoc('').getline(0) == (0, '')
        assert self._getdoc('abc').getline(0) == (3, 'abc')

        doc = self._getdoc('abc\ndef\n')
        assert doc.getline(0) == (4, 'abc\n')
        assert doc.getline(4) == (8, 'def\n')
        assert doc.getline(8) == (8, '')

        doc = self._getdoc('abc\r\ndef')
        assert doc.getline(0) == (5, 'abc\r\n')
        assert doc.getline(5) == (8, 'def')

    def test_style(self):
        assert len(self._getdoc('').styles) == 0
        assert len(self._getdoc('abcde').styles) == 5

        doc = self._getdoc()
        doc.insert(0, 'abcde')
        assert doc.styles.getints(0, len(doc.styles)) == [0] * 5

        doc.insert(1, 'abcde')
        assert doc.styles.getints(0, len(doc.styles)) == [0] * 10

    def test_getnextpos(self):
        doc = self._getdoc("abc\ndef\nghi\njkl\n")
        for i in range(16):
            assert doc.get_nextpos(i) == i + 1

        doc = self._getdoc("abc")
        assert doc.get_nextpos(1) == 2
        assert doc.get_nextpos(2) == 3
        assert doc.get_nextpos(3) == 3

        doc = self._getdoc("abcdef")
        for i in range(6):
            assert doc.get_nextpos(i) == i + 1
        assert doc.get_nextpos(6) == 6

        doc = self._getdoc("abc\r\n")
        assert doc.get_nextpos(3) == 4

        doc = self._getdoc("a\u1dc0\u1dc0\u1dc0\u1dc0b")

        assert doc.get_nextpos(0) == 5
        assert doc.get_nextpos(4) == 5
        assert doc.get_nextpos(5) == 6
        assert doc.get_nextpos(6) == 6

    def test_getprevpos(self):
        doc = self._getdoc("abc\ndef\nghi\njkl\n")
        for i in range(1, 17):
            assert doc.get_prevpos(i) == i - 1

        doc = self._getdoc("abc",)
        assert doc.get_prevpos(0) == 0
        assert doc.get_prevpos(1) == 0
        assert doc.get_prevpos(2) == 1
        assert doc.get_prevpos(3) == 2

        doc = self._getdoc("abcdef")
        for i in range(1, 6):
            assert doc.get_prevpos(i) == i - 1
        assert doc.get_prevpos(0) == 0

        doc = self._getdoc("abc\r\n")
        assert doc.get_prevpos(5) == 4

        doc = self._getdoc("a\u1dc0\u1dc0\u1dc0\u1dc0b")

        assert doc.get_prevpos(0) == 0
        assert doc.get_prevpos(4) == 0
        assert doc.get_prevpos(5) == 0
        assert doc.get_prevpos(6) == 5


class TestMark(kaa_testutils._TestDocBase):

    def test_mark(self):
        doc = self._getdoc('01234567890123456789')
        doc.marks['mark1'] = 0
        doc.marks['mark2'] = 1

        doc.delete(0, 1)

        assert doc.marks['mark1'] == 0
        assert doc.marks['mark2'] == 0

        doc.marks['mark3'] = 3
        doc.insert(0, 'abc')

        assert doc.marks['mark1'] == 0
        assert doc.marks['mark2'] == 0
        assert doc.marks['mark3'] == 6

        doc.delete(0, 3)

        assert doc.marks['mark1'] == 0
        assert doc.marks['mark2'] == 0
        assert doc.marks['mark3'] == 3

    def test_rangemark(self):
        doc = self._getdoc('')
        doc.marks['mark1'] = (0, 0)
        doc.insert(0, 'abcde')
        assert doc.marks['mark1'] == (0, 5)

        doc = self._getdoc('01234567890123456789')
        doc.marks['mark1'] = (1, 3)
        doc.delete(0, 1)
        assert doc.marks['mark1'] == (0, 2)

        doc = self._getdoc('01234567890123456789')
        doc.marks['mark1'] = (1, 3)
        doc.delete(1, 2)
        assert doc.marks['mark1'] == (1, 2)

        doc = self._getdoc('01234567890123456789')
        doc.marks['mark1'] = (1, 3)
        doc.delete(1, 3)
        assert doc.marks['mark1'] == (1, 1)

        doc = self._getdoc('01234567890123456789')
        doc.marks['mark1'] = (1, 3)
        doc.delete(1, 4)
        assert doc.marks['mark1'] == (1, 1)


class TestUndo:

    def test_undo(self):
        undo = document.Undo()
        assert not undo.can_undo()
        assert not undo.can_redo()

        undo.add('action1', 1, arg=1)
        assert undo.can_undo()
        assert not undo.can_redo()

        undo.add('action2', 2, arg=2)
        undo.add('action3', 3, arg=3)

        (action, args, kwargs), = undo.undo()
        assert action == 'action3'
        assert args == (3,)
        assert kwargs == {'arg': 3}

        assert undo.can_undo()
        assert undo.can_redo()
        assert undo.can_redo()

        (action, args, kwargs), = undo.undo()
        assert action == 'action2'
        assert args == (2,)
        assert kwargs == {'arg': 2}

        assert undo.can_undo()
        assert undo.can_redo()

        (action, args, kwargs), = undo.undo()
        assert action == 'action1'
        assert args == (1,)
        assert kwargs == {'arg': 1}

        assert not undo.can_undo()
        assert undo.can_redo()

        (action, args, kwargs), = undo.redo()
        assert action == 'action1'
        assert args == (1,)
        assert kwargs == {'arg': 1}

        assert undo.can_undo()
        assert undo.can_redo()

        (action, args, kwargs), = undo.redo()
        assert action == 'action2'
        assert args == (2,)
        assert kwargs == {'arg': 2}

        assert undo.can_undo()
        assert undo.can_redo()

        (action, args, kwargs), = undo.redo()
        assert action == 'action3'
        assert args == (3,)
        assert kwargs == {'arg': 3}

        assert undo.can_undo()
        assert not undo.can_redo()

    def test_dirty(self):
        undo = document.Undo()
        assert not undo.is_dirty()

        undo.add('action1')
        assert undo.is_dirty()

        undo.add('action2')
        assert undo.is_dirty()
        undo.saved()
        assert not undo.is_dirty()

        undo.add('action3')
        assert undo.is_dirty()

        for r in undo.undo():
            pass
        assert not undo.is_dirty()

    def test_addundo(self):
        undo = document.Undo()
        assert not undo.can_undo()
        assert not undo.can_redo()

        undo.add('action1')
        undo.add('action2')
        undo.saved()

        assert not undo.is_dirty()

        for r in undo.undo():
            pass
        undo.add('action3')

        assert (('action3', (), {}),) == tuple(undo.undo())
        assert (('action1', (), {}),) == tuple(undo.undo())

        assert (('action1', (), {}),) == tuple(undo.redo())
        assert (('action3', (), {}),) == tuple(undo.redo())

        assert undo.is_dirty()

    def test_group(self):
        undo = document.Undo()

        undo.add('1')
        undo.add('2')
        undo.beginblock()
        undo.add('3')
        undo.add('4')
        undo.beginblock()
        undo.add('5')
        undo.add('6')
        undo.endblock()
        undo.add('7')
        undo.endblock()
        undo.saved()
        undo.add('8')

        assert undo.is_dirty()
        assert (('8', (), {}),) == tuple(undo.undo())
        assert not undo.is_dirty()

        assert (('7', (), {}),
                ('6', (), {}),
                ('5', (), {}),
                ('4', (), {}),
                ('3', (), {}),
                ) == tuple(undo.undo())
        assert undo.is_dirty()

        assert (('2', (), {}),) == tuple(undo.undo())
        assert (('1', (), {}),) == tuple(undo.undo())

        assert (('1', (), {}),) == tuple(undo.redo())
        assert (('2', (), {}),) == tuple(undo.redo())
        assert (('3', (), {}),
                ('4', (), {}),
                ('5', (), {}),
                ('6', (), {}),
                ('7', (), {}),
                ) == tuple(undo.redo())
        assert not undo.is_dirty()
        assert (('8', (), {}),) == tuple(undo.redo())
        assert undo.is_dirty()


class TestLineNo(kaa_testutils._TestScreenBase):

    def test_bisect_left(self):
        lineno = document.LineNo()
        lineno.buf.insertints(0, [0, 1, 2, 3, 4, 5])

        assert lineno.bisect_left(0) == 0
        assert lineno.bisect_left(1) == 1
        assert lineno.bisect_left(5) == 5
        assert lineno.bisect_left(6) == 6
        assert lineno.bisect_left(7) == 6

        lineno = document.LineNo()
        assert lineno.bisect_left(0) == 0

    def test_insert(self):
        lineno = document.LineNo()
        lineno.inserted(0, 'abc')
        assert len(lineno.buf) == 0

        lineno.inserted(0, 'abc\n')
        assert lineno.buf.getints(0, len(lineno.buf)) == [3]

        lineno.inserted(0, 'abc\n')
        assert lineno.buf.getints(0, len(lineno.buf)) == [3, 7]

        lineno.inserted(0, '\na\nab\n')
        assert lineno.buf.getints(0, len(lineno.buf)) == [0, 2, 5, 9, 13]

        lineno.inserted(14, 'a\nb')
        assert lineno.buf.getints(0, len(lineno.buf)) == [0, 2, 5, 9, 13, 15]

    def test_delete(self):
        lineno = document.LineNo()
        lineno.inserted(0, 'abc\n1def\n23ghi\n')

        lineno.deleted(0, 2)
        assert lineno.buf.getints(0, len(lineno.buf)) == [1, 6, 12]

        lineno.deleted(4, 8)
        assert lineno.buf.getints(0, len(lineno.buf)) == [1, 8]

        lineno = document.LineNo()
        lineno.inserted(0, 'abcdefg')

        lineno.deleted(0, 2)
        assert lineno.buf.getints(0, len(lineno.buf)) == []

        lineno = document.LineNo()
        lineno.inserted(0, 'abc')
        lineno.deleted(0, 100)
        assert lineno.buf.getints(0, len(lineno.buf)) == []

        lineno = document.LineNo()
        lineno.inserted(0, 'abcde')
        lineno.deleted(0, 2)
        assert lineno.buf.getints(0, len(lineno.buf)) == []

        lineno = document.LineNo()
        lineno.inserted(0, 'abcde\nfghi')
        lineno.deleted(0, 2)
        assert lineno.buf.getints(0, len(lineno.buf)) == [3]

        lineno = document.LineNo()
        lineno.inserted(0, 'abcde\nfghi')
        lineno.deleted(6, 8)
        assert lineno.buf.getints(0, len(lineno.buf)) == [5]

        lineno = document.LineNo()
        lineno.inserted(0, 'abcde\nfghi')
        lineno.deleted(3, 8)
        assert lineno.buf.getints(0, len(lineno.buf)) == []

        lineno = document.LineNo()
        lineno.inserted(0, 'abc\ndef')
        lineno.deleted(0, 100)
        assert lineno.buf.getints(0, len(lineno.buf)) == []



import os
import sys
from setuptools import setup, find_packages, Extension


if sys.version_info[:2] < (3, 4):
    raise RuntimeError('''Kaa requires Python 3.4 or later, with ncursesw support. 
Please see https://pypi.python.org/pypi/kaaedit/#setup to install.
''')

try:
    from Cython.Distutils import build_ext
    cmdclass = {'build_ext': build_ext}

    ext = Extension("_gappedbuf",
                    ["_gappedbuf/_gappedbuf.pyx",
                     "_gappedbufre/_sre.c"])

except ImportError:
    cmdclass = {}
    ext = Extension("_gappedbuf",
                    ["_gappedbuf/_gappedbuf.c",
                     "_gappedbufre/_sre.c"])


def read(fname):
    return open(
        os.path.join(os.path.dirname(__file__), fname)).read()

bdist_wheel = {}

if sys.platform == 'linux':
    bdist_wheel['plat_name'] = 'manylinux2014_x86_64'

setup(
    cmdclass=cmdclass,
    name="kaaedit",
    version="0.54.0",
    description='kaa - console text editor.',
    url='https://github.com/kaaedit/kaa',
    author='Atsuo Ishimoto',
    author_email='ishimoto@gembook.org',
    long_description=read('README.rst'),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Development Status :: 3 - Alpha",
        "Topic :: Text Editors",
        "Topic :: Software Development :: Debuggers",
        "Environment :: Console :: Curses",
        "License :: OSI Approved :: MIT License", ],
    license='MIT License',
    install_requires=['curses_ex', 'pyjf3', 'setproctitle', 'gitpython', 
        'kaadbg >= 0.3.0'],
    packages=find_packages(),
    ext_modules=[ext],
    options={'bdist_wheel': bdist_wheel},
    entry_points={
        'console_scripts': [
            'kaa = kaa.cui.main:run',
        ], }
)



import curses
import curses_ex
import io
import os
from kaa.keyboard import *

KAA_KEYCODE_FROM = 1000

BUTTONNAMES = {
    curses.BUTTON1_CLICKED: "BUTTON1_CLICKED",
    curses.BUTTON1_DOUBLE_CLICKED: "BUTTON1_DOUBLE_CLICKED",
    curses.BUTTON1_PRESSED: "BUTTON1_PRESSED",
    curses.BUTTON1_RELEASED: "BUTTON1_RELEASED",
    curses.BUTTON1_TRIPLE_CLICKED: "BUTTON1_TRIPLE_CLICKED",
    curses.BUTTON2_CLICKED: "BUTTON2_CLICKED",
    curses.BUTTON2_DOUBLE_CLICKED: "BUTTON2_DOUBLE_CLICKED",
    curses.BUTTON2_PRESSED: "BUTTON2_PRESSED",
    curses.BUTTON2_RELEASED: "BUTTON2_RELEASED",
    curses.BUTTON2_TRIPLE_CLICKED: "BUTTON2_TRIPLE_CLICKED",
    curses.BUTTON3_CLICKED: "BUTTON3_CLICKED",
    curses.BUTTON3_DOUBLE_CLICKED: "BUTTON3_DOUBLE_CLICKED",
    curses.BUTTON3_PRESSED: "BUTTON3_PRESSED",
    curses.BUTTON3_RELEASED: "BUTTON3_RELEASED",
    curses.BUTTON3_TRIPLE_CLICKED: "BUTTON3_TRIPLE_CLICKED",
    curses.BUTTON4_CLICKED: "BUTTON4_CLICKED",
    curses.BUTTON4_DOUBLE_CLICKED: "BUTTON4_DOUBLE_CLICKED",
    curses.BUTTON4_PRESSED: "BUTTON4_PRESSED",
    curses.BUTTON4_RELEASED: "BUTTON4_RELEASED",
    curses.BUTTON4_TRIPLE_CLICKED: "BUTTON4_TRIPLE_CLICKED",
    curses.BUTTON_ALT: "BUTTON_ALT",
    curses.BUTTON_CTRL: "BUTTON_CTRL",
    curses.BUTTON_SHIFT: "BUTTON_SHIFT",
}


class KeyEvent:

    """ Stores information of keyboard input event.

    Attributes:
        key -- key code returned by curses.
    """
    KEY_TRANSLATE = {
        curses.KEY_BACKSPACE: '\x7f'
    }

    def __init__(self, key, has_trailing_char):
        """

        Auguments:
            key -- key code returned by curses.

        """

        self.key = self.KEY_TRANSLATE.get(key, key)
        self.has_trailing_char = has_trailing_char

    def __repr__(self):
        c = self.key if isinstance(self.key, int) else ord(self.key)
        name = str(curses.keyname(c), 'ascii', 'replace')
        return '<key: {!r} {}'.format(name, super().__repr__())

    def unget(self):
        curses.unget_wch(self.key)


class MouseEvent:

    """ Stores information of mouse input event.

    Attributes:
        x, y -- Event coordinates
        code -- Type of event
        shift -- True if shift key was pressed
        ctrl -- True if control key was pressed
        alt -- True if alt key was pressed
    """

    def __init__(self, mouseid, x, y, z, bstate):
        self.mouseid = mouseid
        self.x = x
        self.y = y
        self.z = z
        self.bstate = bstate

        self.code = bstate & ~(curses.BUTTON_SHIFT | curses.BUTTON_CTRL |
                               curses.BUTTON_ALT)

        self.shift = bstate & curses.BUTTON_SHIFT
        self.ctrl = bstate & curses.BUTTON_CTRL
        self.alt = bstate & curses.BUTTON_ALT

    @property
    def button1clicked(self):
        """True if code is BUTTON1_CLICKED"""

        return self.code == curses.BUTTON1_CLICKED

    @property
    def button2clicked(self):
        """True if code is BUTTONR2_CLICKED"""

        return self.code == curses.BUTTON2_CLICKED

    @property
    def button3clicked(self):
        """True if code is BUTTON3_CLICKED"""

        return self.code == curses.BUTTON3_CLICKED

    def __repr__(self):
        name = BUTTONNAMES.get(self.code, self.code)
        name = hex(name) if isinstance(name, int) else name

        mods = "".join((('M-' if self.alt else ''),
                        ('^' if self.ctrl else ''),
                        ('SHIFT+' if self.shift else '')))

        return '<mouse: ({},{}) - {}{} {}>'.format(
            self.x, self.y, mods, name, super().__repr__())

    def unget(self):
        curses.ungetmouse(self.mouseid, self.x, self.y, self.z, self.bstate)


KEYNAME_CURSES = {
    backspace: '\x7f',
    delete: curses.KEY_DC,
    pageup: curses.KEY_PPAGE,
    pagedown: curses.KEY_NPAGE,
    left: curses.KEY_LEFT,
    right: curses.KEY_RIGHT,
    up: curses.KEY_UP,
    down: curses.KEY_DOWN,
    home: curses.KEY_HOME,
    end: curses.KEY_END,
    tab: '\t',
    esc: '\x1b',
    f1: curses.KEY_F1,
    f2: curses.KEY_F2,
    f3: curses.KEY_F3,
    f4: curses.KEY_F4,
    f5: curses.KEY_F5,
    f6: curses.KEY_F6,
    f7: curses.KEY_F7,
    f8: curses.KEY_F8,
    f9: curses.KEY_F9,
    f10: curses.KEY_F10,
    f11: curses.KEY_F11,
    f12: curses.KEY_F12,
    f13: curses.KEY_F13,
    f14: curses.KEY_F14,
    f15: curses.KEY_F15,
    f16: curses.KEY_F16,
    f17: curses.KEY_F17,
    f18: curses.KEY_F18,
    f19: curses.KEY_F19,
    f20: curses.KEY_F20,
    f21: curses.KEY_F21,
    f22: curses.KEY_F22,
    f23: curses.KEY_F23,
    f24: curses.KEY_F24,
}


KEYNAME_CURSES_SHIFT = {
    tab: curses.KEY_BTAB,
    delete: curses.KEY_SDC,
    home: curses.KEY_SHOME,
    end: curses.KEY_SEND,
    left: curses.KEY_SLEFT,
    right: curses.KEY_SRIGHT,
    up: curses.KEY_SR,
    down: curses.KEY_SF
}

KEYNAME_TO_CODE = {
    # (ctrl, shift, key) : keycode
    (0, 0, backspace): '\x7f',
    (0, 0, delete): curses.KEY_DC,
    (0, 0, pageup): curses.KEY_PPAGE,
    (0, 0, pagedown): curses.KEY_NPAGE,
    (0, 0, left): curses.KEY_LEFT,
    (0, 0, right): curses.KEY_RIGHT,
    (0, 0, up): curses.KEY_UP,
    (0, 0, down): curses.KEY_DOWN,
    (0, 0, home): curses.KEY_HOME,
    (0, 0, end): curses.KEY_END,
    (0, 0, tab): '\t',
    (0, 0, esc): '\x1b',
    (0, 0, f1): curses.KEY_F1,
    (0, 0, f2): curses.KEY_F2,
    (0, 0, f3): curses.KEY_F3,
    (0, 0, f4): curses.KEY_F4,
    (0, 0, f5): curses.KEY_F5,
    (0, 0, f6): curses.KEY_F6,
    (0, 0, f7): curses.KEY_F7,
    (0, 0, f8): curses.KEY_F8,
    (0, 0, f9): curses.KEY_F9,
    (0, 0, f10): curses.KEY_F10,
    (0, 0, f11): curses.KEY_F11,
    (0, 0, f12): curses.KEY_F12,
    (0, 0, f13): curses.KEY_F13,
    (0, 0, f14): curses.KEY_F14,
    (0, 0, f15): curses.KEY_F15,
    (0, 0, f16): curses.KEY_F16,
    (0, 0, f17): curses.KEY_F17,
    (0, 0, f18): curses.KEY_F18,
    (0, 0, f19): curses.KEY_F19,
    (0, 0, f20): curses.KEY_F20,
    (0, 0, f21): curses.KEY_F21,
    (0, 0, f22): curses.KEY_F22,
    (0, 0, f23): curses.KEY_F23,
    (0, 0, f24): curses.KEY_F24,

    (0, 1, tab): curses.KEY_BTAB,
    (0, 1, delete): curses.KEY_SDC,
    (0, 1, home): curses.KEY_SHOME,
    (0, 1, end): curses.KEY_SEND,
    (0, 1, left): curses.KEY_SLEFT,
    (0, 1, right): curses.KEY_SRIGHT,
    (0, 1, up): curses.KEY_SR,
    (0, 1, down): curses.KEY_SF,
    (0, 1, pagedown): curses.KEY_SNEXT,
    (0, 1, pageup): curses.KEY_SPREVIOUS,
}


def keyfromname(key, ctrl, shift):
    return KEYNAME_TO_CODE.get((ctrl, shift, key))


def convert_registered_key(key):
    if key in REGISTERD_SEQUENCE:
        return REGISTERD_SEQUENCE[key]
    else:
        return [key]

# works for xterm?
CAPNAME_TO_KEY = {
    # capname: (ctrl, shift, key),
    'kHOM5': (1, 0, home),
    'kHOM6': (1, 1, home),
    'kEND5': (1, 0, end),
    'kEND6': (1, 1, end),
    'kLFT5': (1, 0, left),
    'kLFT6': (1, 1, left),
    'kRIT5': (1, 0, right),
    'kRIT6': (1, 1, right),
    'kUP5': (1, 0, up),
    'kUP6': (1, 1, up),
    'kDN5': (1, 0, down),
    'kDN6': (1, 1, down),
    'kPRV5': (1, 0, pageup),
    'kPRV6': (1, 1, pageup),
    'kNXT5': (1, 0, pagedown),
    'kNXT6': (1, 1, pagedown),
    'kDC5': (1, 0, delete),
    'kDC6': (1, 1, delete),
}

# http://pubs.opengroup.org/onlinepubs/7908799/xcurses/terminfo.html
CAPNAME_TO_CODE = {
    'cuf1': curses.KEY_RIGHT,
    'kcuf1': curses.KEY_RIGHT,
    'kRIT': curses.KEY_SRIGHT,
    'kcub1': curses.KEY_LEFT,
    'kcuu1': curses.KEY_UP,
    'kri': curses.KEY_SR,
    'kcud1': curses.KEY_DOWN,
    'kind': curses.KEY_SF,
    'kDN': curses.KEY_SF,
    'khome': curses.KEY_HOME,
    'kHOM': curses.KEY_SHOME,
    'kend': curses.KEY_END,
    'kEND': curses.KEY_SEND,
    'kpp': curses.KEY_PPAGE,
    'knp': curses.KEY_NPAGE,
}

# Terminals like iTerm2 sends following codes for alt+cursor keys
KNOWN_KEY_SEQUENCE = {
    b'\x1b\x1b[A': ['\x1b', curses.KEY_UP],
    b'\x1b\x1b[B': ['\x1b', curses.KEY_DOWN],
    b'\x1b\x1b[C': ['\x1b', curses.KEY_RIGHT],
    b'\x1b\x1b[D': ['\x1b', curses.KEY_LEFT],
}

REGISTERD_SEQUENCE = {}


def init(conf):
    # Experimental code to get keycode from cap name
    for i in range(512, 1024):
        name = str(curses.keyname(i), 'utf-8', 'replace')
        if name in CAPNAME_TO_KEY:
            KEYNAME_TO_CODE[CAPNAME_TO_KEY[name]] = i
            CAPNAME_TO_CODE[name] = i

    result = io.StringIO()
    err = False

    reg_code = KAA_KEYCODE_FROM
    # register ESC + seq
    registered_seqs = set()
    for name, keycode in CAPNAME_TO_CODE.items():
        seq = curses.tigetstr(name)
        if seq:
            seq = b'\x1b' + seq
            try:
                curses_ex.define_key(seq, reg_code)
                print('OK: ', name, keycode, reg_code, seq, file=result)
            except Exception:
                print('NG: ', name, keycode, reg_code, seq, file=result)
                continue

            REGISTERD_SEQUENCE[reg_code] = ['\x1b', keycode]
            reg_code += 1
            registered_seqs.add(seq)

    for seq, keycode in KNOWN_KEY_SEQUENCE.items():
        if seq not in registered_seqs:
            try:
                curses_ex.define_key(seq, reg_code)
                print('OK: ', keycode, reg_code, seq, file=result)
            except Exception:
                print('NG: ', keycode, reg_code, seq, file=result)
                continue

            REGISTERD_SEQUENCE[reg_code] = keycode
            reg_code += 1

    logname = os.path.join(conf.LOGDIR, 'DEFINE_KEY.LOG')
    with open(logname, 'w') as f:
        f.write(result.getvalue())



import os
import bisect
import sys
import threading
import socket
import tempfile
from git import Repo
from git import BaseIndexEntry
from pathlib import Path
import kaa
from kaa.keyboard import *
from kaa.theme import Style
from kaa.filetype.default import defaultmode
from kaa import document
from kaa.filetype.default import keybind
from kaa.command import commandid
from kaa.ui.git import commitdlgmode
from kaa.ui.viewdiff import viewdiffmode
from kaa.ui.msgbox import msgboxmode
from . import gitrepo

GitStatusTheme = {
    'basic': [
        Style('git-header', 'Cyan', None),
        Style('git-button', 'Orange', 'Base02'),
        Style('git-untracked', 'Orange', None),
        Style('git-not-staged', 'Red', None),
        Style('git-staged', 'Green', None),
    ],
}


status_keys = {
    right: 'git.status.next',
    down: 'git.status.next',
    left: 'git.status.prev',
    up: 'git.status.prev',
    pagedown: 'cursor.pagedown',
    pageup: 'cursor.pageup',

    ' ': 'git.status.press',
    '\r': 'git.status.open',
    '\n': 'git.status.open',
    'r': 'git.status.refresh',
    'c': 'git.commit',
    'a': 'git.add',
    'u': 'git.unstage',
    'd': 'git.status.diff',
}


class GitStatusMode(defaultmode.DefaultMode):
    # todo: rename key/command names
    DOCUMENT_MODE = False
    USE_UNDO = False
    DEFAULT_MENU_MESSAGE = 'r:refresh a:add u:unstage d:diff c:commit'

    KEY_BINDS = [
        keybind.app_keys,
        keybind.search_command_keys,
    ]

    def init_keybind(self):
        self.register_keys(self.keybind, self.KEY_BINDS)
        self.keybind.add_keybind(status_keys)

    def init_themes(self):
        super().init_themes()
        self.themes.append(GitStatusTheme)

    def on_str(self, wnd, s, overwrite=False):
        # does nothing
        pass

    def on_esc_pressed(self, wnd, event):
        is_available, command = self.get_command('file.close')
        command(wnd)

    def on_add_window(self, wnd):
        super().on_add_window(wnd)
        self.file_next(wnd)

    def _get_marks(self):
        return sorted(((v[0], v[1], k) for k, v in self.document.marks.items()), key=lambda o: o[1])

    def _get_next_mark(self, pos):
        marks = self._get_marks()
        starts = [v[0] for v in marks]
        start = idx = bisect.bisect_right(starts, pos)

        for idx in range(start, len(marks)):
            yield marks[idx]

        idx = 0
        for idx in range(0, start):
            yield marks[idx]

    def _get_prev_mark(self, pos):
        marks = self._get_marks()
        starts = [v[0] for v in marks]
        start = idx = bisect.bisect_left(starts, pos)

        for idx in range(start - 1, -1, -1):
            yield marks[idx]

        for idx in range(len(marks) - 1, start - 1, -1):
            yield marks[idx]

    def _get_mark(self, pos):
        for f, t, name in self._get_marks():
            if f <= pos < t:
                return f, t, name

    def _get_file_mark(self, pos):
        ret = self._get_mark(pos)
        if ret:
            if ret[2].startswith(('s_', 'n_', 'u_')):
                return ret

    @commandid('git.status.open')
    def open_file(self, wnd):
        mark = self._get_file_mark(wnd.cursor.pos)
        if not mark:
            return

        is_available, command = self.get_command('file.close')
        command(wnd)

        f, t, name = mark

        is_available, command = self.get_command('file.open')
        command(None,
                filename=os.path.join(self._repo.working_dir, name[2:]))

    def _show_diff(self, s):
        doc = document.Document()
        doc.append(s)
        mode = viewdiffmode.ViewDiffMode()
        doc.setmode(mode)

        kaa.app.show_dialog(doc)

    @commandid('git.status.diff')
    def diff_file(self, wnd):
        mark = self._get_file_mark(wnd.cursor.pos)
        if not mark:
            return
        f, t, name = mark
        if name[:2] == 'n_':
            d = self._repo.git.diff(name[2:])
            self._show_diff(d)
        elif name[:2] == 's_':
            d = self._repo.git.diff('--cached', name[2:])
            self._show_diff(d)

    @commandid('git.status.next')
    def file_next(self, wnd):
        for mark in self._get_next_mark(wnd.cursor.pos):
            wnd.cursor.setpos(mark[0])
            break

    @commandid('git.status.prev')
    def file_prev(self, wnd):
        for mark in self._get_prev_mark(wnd.cursor.pos):
            wnd.cursor.setpos(mark[0])
            break

    @commandid('git.status.press')
    def status_press(self, wnd):
        mark = self._get_mark(wnd.cursor.pos)
        if not mark:
            return
        f, t, name = mark
        if name == 'b_refresh':
            self._refresh(wnd)

    @commandid('git.status.refresh')
    def status_refresh(self, wnd):
        self._refresh(wnd)

    @commandid('git.add')
    def git_add(self, wnd):
        mark = self._get_file_mark(wnd.cursor.pos)
        if not mark:
            return
        f, t, name = mark
        self._repo.git.add([name[2:]])

        self._refresh(wnd)

    @commandid('git.unstage')
    def git_unstage(self, wnd):
        mark = self._get_file_mark(wnd.cursor.pos)
        if not mark:
            return
        f, t, name = mark

        if self._repo.head.is_valid():
            self._repo.git.reset('HEAD', [name[2:]])
        else:
            self._repo.git.rm('--cached', [name[2:]])

        self._refresh(wnd)

    @commandid('git.commit')
    def git_commit(self, wnd):

        fname = tempfile.mktemp()
        if os.path.exists(fname):
            os.unlink(fname)

        git_edit = '{exe} -m kaa.ui.git.git_editor '.format(exe=sys.executable)
        self._repo.git.update_environment(
            KAA_SOCKNAME=fname, GIT_EDITOR=git_edit)

        git_result = None

        def f():
            nonlocal git_result
            try:
                code, msg1, msg2 = git_result = self._repo.git.commit(
                    with_extended_output=True)
                # todo: check gitpython to know what they are...
                git_result = msg1 + msg2
            except Exception as e:
                git_result = e

            s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            try:
                s.connect(fname)
                s.send(b'finished\n')
                s.close()
            except FileNotFoundError:
                pass

        s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        try:
            s.bind(fname)
            s.listen()

            t = threading.Thread(target=f, daemon=True)
            t.start()

            conn, addr = s.accept()
            if git_result:
                kaa.app.messagebar.set_message(
                    ' '.join(str(git_result).split('\n')))

            def callback_commit():
                t.join()
                if git_result:
                    msg = str(git_result)
                    msgboxmode.MsgBoxMode.show_msgbox(
                        msg, ['&Ok'], lambda c: self.status_refresh(wnd), [
                            '\r', '\n'],
                        border=True)
                else:
                    self.status_refresh(wnd)

            recv = conn.recv(4096)
            if recv == b'ok\n':
                filename = os.path.join(self._repo.git_dir, 'COMMIT_EDITMSG')
                doc = kaa.app.storage.openfile(
                    filename, nohist=True, filemustexists=True)
                s = doc.gettext(0, doc.endpos())
                doc.close()

                doc = commitdlgmode.CommitDialogMode.build(s)
                # todo: bad interface
                doc.mode.commitmsg = filename
                doc.mode.conn = conn

                doc.mode.commit_callback = callback_commit
                kaa.app.show_dialog(doc)
            else:
                callback_commit()

        finally:
            if os.path.exists(fname):
                os.unlink(fname)

    def _add_new_file(self, indent, style, name, mark_prefix):
        mark = mark_prefix + name
        if mark in self.document.marks:
            return

        f = self.document.append(indent+'new file:    ', style)
        t = self.document.append(name, style)
        self.document.marks[mark] = (f, t)
        return True

    def _add_diff(self, indent, d, style, mark_prefix):
        if d.new_file:
            return self._add_new_file(indent, style, d.b_path, mark_prefix)

        elif d.deleted_file:
            mark = mark_prefix + d.b_path
            if mark in self.document.marks:
                return

            f = self.document.append(indent + 'deleted:     ', style)
            t = self.document.append(d.b_path, style)
            self.document.marks[mark] = (f, t)

        elif d.renamed:
            mark_from = mark_prefix + d.rename_from
            if mark_from in self.document.marks:
                return

            mark_to = mark_prefix + d.rename_to
            if mark_to in self.document.marks:
                return

            f = self.document.append(indent + 'renamed:     ', style)
            t = self.document.append(d.rename_from, style)
            self.document.marks[mark_from] = (f, t)

            f = self.document.append(' -> ', style)
            t = self.document.append(d.rename_to, style)
            self.document.marks[mark_to] = (f, t)

        else:
            mark = mark_prefix + d.b_path
            if mark in self.document.marks:
                return

            f = self.document.append(indent + 'modified:    ', style)
            t = self.document.append(d.b_path, style)
            self.document.marks[mark_prefix + d.b_path] = (f, t)

        return True

    def _refresh(self, wnd):
        if wnd:
            lineno = self.document.buf.lineno.lineno(wnd.cursor.pos)

        # clear
        self.document.marks.clear()
        self.document.delete(0, self.document.endpos())
        self.document.set_title('<git status>')

        untracked_files = list(self._repo.untracked_files)

        # todo: define lock_mark() method
        self.document.marks.locked = True
        indent = '    '
        try:
            style_header = self.get_styleid('git-header')
            style_button = self.get_styleid('git-button')
            style_staged = self.get_styleid('git-staged')
            style_not_staged = self.get_styleid('git-not-staged')
            style_untracked = self.get_styleid('git-untracked')

            try:
                branch = 'On branch {repo.active_branch.name}'.format(repo=self._repo)
            except TypeError:
                branch = '(no branch)'

            f = self.document.append(branch+'\n\n', style_header)
            t = self.document.append('<< Refresh >>', style_button)
            self.document.marks['b_refresh'] = (f, t)

            self.document.append('\n\n')

            # unmerged
            unmerged = {}
            if self._repo.head.is_valid():
                d = self._repo.index.unmerged_blobs()
                for path, blobs in d.items():
                    stagemask = 0
                    for stage, blob in blobs:
                        stagemask |= (1 << (stage-1))   # see man git-merge, wt-status.c: unmerged_mask()

                    unmerged[path] = stagemask

            
            # add staged files
            unmerged_diffs = []
            self.document.append('Changes to be committed:\n\n', style_header)
            if self._repo.head.is_valid():
                d = self._repo.head.commit.diff()
                for c in d:
                    if c.b_path in unmerged:
                        unmerged_diffs.append(c)
                        continue

                    if self._add_diff(indent, c, style_staged, 's_'):
                        self.document.append('\n')
            else:
                files = [name for name, state in self._repo.index.entries]
                files.sort()
                for file in files:
                    if self._add_new_file(indent, style_staged, file, 's_'):
                        self.document.append('\n')

            # add unmerged files
            if unmerged_diffs:
                self.document.append('\nUnmerged paths:\n\n', style_header)
                for c in unmerged_diffs:
                    if self._add_diff(indent, c, style_staged, 'n_'):
                        self.document.append('\n')

            # add not staged files
            self.document.append(
                '\nChanges not staged for commit:\n\n', style_header)

            d = self._repo.index.diff(None)
            for c in d:
                if self._add_diff(indent, c, style_not_staged, 'n_'):
                    self.document.append('\n')

            # add untracked files
            self.document.append('\n\nUntracked fies:\n\n', style_header)
            for f in untracked_files:
                self.document.append(indent)
                posfrom = self.document.endpos()
                self.document.append(f, style_untracked)
                posto = self.document.endpos()

                self.document.marks['u_' + f] = (posfrom, posto)
                self.document.append('\n')

        finally:
            self.document.marks.locked = False

        if wnd:
            pos = self.document.get_lineno_pos(lineno)
            tol = wnd.document.gettol(pos)
            for f, t, name in self._get_marks():
                if tol <= f:
                    break
            wnd.cursor.setpos(f)

        kaa.app.messagebar.set_message(self.DEFAULT_MENU_MESSAGE)

    def show_status(self, repo):
        self._repo = repo
        self._refresh(None)


def show_git_status(d):
    repo = gitrepo.open_repo(d)

    doc = document.Document()
    mode = GitStatusMode()
    doc.setmode(mode)

    mode.show_status(repo)
    kaa.app.show_doc(doc)



import os


class FileTypeInfo:
    FILE_EXT = {}

    @classmethod
    def select_mode(cls, filename):
        ext = os.path.splitext(filename)[1].lower()
        if ext in cls.FILE_EXT:
            return cls.get_modetype()

    @classmethod
    def get_modetype(cls):
        assert 0




KAA_DIR = '~/.kaa/'
KAA_SCRIPT = '__kaa__.py'
LOGDIR = 'log'
HISTDIR = 'history'
HIST_DBNAME = 'hist.db'
SPELLCHECKER_PWL = 'spell_pwl'

NEWLINES = ['auto', 'dos', 'unix']
NEWLINE_CHARS = {
    'auto': None,
    'dos': '\r\n',
    'unix': '\n'
}



from contextlib import ExitStack
from unittest.mock import patch
from kaa import cursor

import kaa_testutils


class TestCursor(kaa_testutils._TestScreenBase):

    def _getcursor(self, s, width=10, height=10):
        wnd = self._getwnd(s, width, height)
        return cursor.Cursor(wnd)

    def test_up(self):
        cursor = self._getcursor("abc\ndef\nghi\n012\n345\n678\n", height=3)
        cursor.preferred_col = 2

        cursor.setpos(10)
        assert cursor.up()
        assert cursor.pos == 6

        assert cursor.up()
        assert cursor.pos == 2

        assert not cursor.up()
        assert cursor.pos == 2

        cursor.setpos(24)
        assert cursor.up()
        assert cursor.pos == 22

        assert cursor.up()
        assert cursor.pos == 18

    def test_down(self):
        cursor = self._getcursor("abc\ndef\nghi\n012\n345\n678\n", height=3)
        cursor.preferred_col = 10

        cursor.setpos(0)
        assert cursor.down()
        assert cursor.pos == 7

        assert cursor.down()
        assert cursor.pos == 11

        cursor.setpos(16)
        assert cursor.down()
        assert cursor.pos == 23

        assert cursor.down()
        assert cursor.pos == 24

    def test_right(self):
        s = "abc\ndef\nghi\n012\n345\n678\n"
        cursor = self._getcursor(s, height=1)
        cursor.setpos(0)
        for i in range(len(s)):
            cursor.right()
            assert cursor.pos == i + 1

    def test_rightword(self):
        s = 'abc あいうえおカキクケコ'
        cursor = self._getcursor(s, height=1, width=30)
        cursor.setpos(0)

        cursor.right(word=True)
        assert cursor.pos == 4

        cursor.right(word=True)
        assert cursor.pos == 9

        cursor.right(word=True)
        assert cursor.pos == 14

    def test_left(self):
        s = "abc\ndef\nghi\n012\n345\n678\n"
        cursor = self._getcursor(s, height=1)
        cursor.setpos(len(s))
        for i in range(len(s) - 1, -1, -1):
            cursor.left()
            assert cursor.pos == i

    def test_leftword(self):
        s = 'abc あいうえおカキクケコ'
        cursor = self._getcursor(s, height=1, width=30)
        cursor.setpos(len(s))

        cursor.left(word=True)
        assert cursor.pos == 9

        cursor.left(word=True)
        assert cursor.pos == 4

        cursor.left(word=True)
        assert cursor.pos == 0

    def test_pagedown(self):
        s = "abc\ndef\nghi\n012\n345\n678\n"
        cursor = self._getcursor(s, height=3)
        cursor.setpos(0)
        cursor.preferred_col = 1

        cursor.pagedown()
        assert cursor.pos == 9

        cursor.pagedown()
        assert cursor.pos == 17

        cursor.pagedown()
        assert cursor.pos == 24

    def test_pageup(self):
        s = "abc\ndef\nghi\n012\n345\n678\n"
        cursor = self._getcursor(s, height=3)
        cursor.setpos(24)
        cursor.preferred_col = 1

        cursor.pageup()
        assert cursor.pos == 17

        cursor.pageup()
        assert cursor.pos == 9

        cursor.pageup()
        assert cursor.pos == 5



from unittest.mock import patch, Mock

import kaa
from kaa.ui.mainmenu import menumode
import kaa_testutils

kaa.app.show_dialog = Mock()


class TestMenuMode(kaa_testutils._TestDocBase):

    def test_msbbox(self):
        kaa.app.DEFAULT_THEME = 'basic'

        target = kaa_testutils._TestScreenBase()._getwnd('')
        doc = menumode.MenuMode.show_menu(target, 'MAIN')
        doc.mode.on_str(None, 'f')

        name, args, kwargs = kaa.app.show_dialog.mock_calls[-1]
        assert args[0].mode.itemname == 'FILE'



import kaa_testutils
from kaa.commands import editorcommand
import kaa.macro
import kaa.command
from kaa.command import commandid, norec


class TestMacro(kaa_testutils._TestScreenBase):

    def test_toggle_record(self):
        macro = kaa.macro.Macro()
        assert not macro.is_recording()

        macro.toggle_record()
        assert macro.is_recording()

        macro.toggle_record()
        assert not macro.is_recording()

    def test_record(self):

        macro = kaa.macro.Macro()
        macro.toggle_record()

        @commandid('command1')
        def f():
            pass

        macro.record(1, f)
        assert macro.get_commands() == [
            (kaa.macro.rec_command, 1, 'command1', (), {})]

        @commandid('command2')
        @norec
        def g():
            pass

        macro.record(1, g)
        assert macro.get_commands() == [
            (kaa.macro.rec_command, 1, 'command1', (), {})]

    def test_run(self):
        wnd = self._getwnd("abc\ndef\n")

        macro = kaa.macro.Macro()
        macro.toggle_record()

        mode = wnd.document.mode
        macro.record(1, mode.get_command('cursor.right')[1])
        macro.record(1, mode.get_command('cursor.right')[1])
        macro.record(1, mode.get_command('cursor.right')[1])
        macro.record(1, mode.get_command('cursor.left')[1])
        macro.toggle_record()

        macro.run(wnd)

        assert wnd.cursor.pos == 2

        macro.toggle_record()

        macro.record_string('123', False)
        macro.record_string('45', False)
        macro.toggle_record()

        macro.run(wnd)

        assert wnd.cursor.pos == 7
        assert wnd.document.gettext(0, 7) == 'ab12345'



import difflib
import kaa
from kaa import document
from kaa.filetype.diff import diffmode
from kaa.filetype.default import keybind
from kaa.ui.dialog import dialogmode


class ViewDiffMode(dialogmode.DialogMode):
    MODENAME = 'ViewDiff'
    DOCUMENT_MODE = False
    USE_UNDO = False

    KEY_BINDS = [
        keybind.cursor_keys,
        keybind.edit_command_keys,
        keybind.emacs_keys,
        keybind.macro_command_keys,
    ]

    tokenizer = diffmode.make_tokenizer()
    callback = None

    def init_keybind(self):
        super().init_keybind()

        self.register_keys(self.keybind, self.KEY_BINDS)

    def init_themes(self):
        super().init_themes()
        self.themes.append(diffmode.DiffThemes)

    def is_cursor_visible(self):
        return 1   # hide cursor

    def on_str(self, wnd, s, overwrite=False):
        # does nothing
        pass

    def on_esc_pressed(self, wnd, event):
        popup = wnd.get_label('popup')
        if popup:
            popup.destroy()
            kaa.app.messagebar.set_message("")
            if self.callback:
                self.callback()


def view_doc_diff(curdoc, callback=None):
    orig = kaa.app.storage.openfile(
        curdoc.fileinfo.fullpathname,
        curdoc.fileinfo.encoding,
        curdoc.fileinfo.newline,
        nohist=True)

    org_lines = list(orig.iterlines(0))
    orig.close()

    cur_lines = list(curdoc.iterlines(0))
    diff = ''.join(difflib.unified_diff(org_lines, cur_lines,
                                        curdoc.fileinfo.fullpathname, '(buffer)'))

    doc = document.Document()
    doc.insert(0, diff)

    mode = ViewDiffMode()
    mode.callback = callback
    doc.setmode(mode)

    kaa.app.show_dialog(doc)


def show_diff(diff):
    doc = document.Document()
    doc.insert(0, diff)

    mode = ViewDiffMode()
    doc.setmode(mode)

    kaa.app.show_dialog(doc)



import datetime, time
import kaa
from kaa.keyboard import *
from kaa.theme import Style
from kaa.filetype.default import defaultmode
from kaa import document
from kaa.filetype.default import keybind
from kaa.command import commandid
from kaa.ui.git import commitdlgmode
from kaa.ui.viewdiff import viewdiffmode
from kaa.ui.msgbox import msgboxmode
from kaa.ui.dialog import dialogmode
from kaa.ui.viewdiff import viewdiffmode

from . import gitrepo



TZ = datetime.timezone(datetime.timedelta(seconds=time.timezone*-1), 
    time.tzname[0])

GitLogTheme = {
    'basic': [
        Style('default', 'White', 'Base01', False, False),
        Style('git-log-header', 'Orange', 'Base02', fillrow=True),

        # todo: right-button with rjust should be defined in another name.
        Style('right-button', 'Base3', 'Base02', rjust=True, nowrap=True),
        Style('right-button.checked', 'Base3', 'Orange', rjust=True,
              nowrap=True),
        Style('right-button.shortcut', 'Base3', 'Base02', underline=True,
              rjust=True, nowrap=True),
        Style('right-button.shortcut.checked', 'Base3', 'Orange',
              underline=True, rjust=True, nowrap=True),

    ],
}


log_keys = {
    tab: 'git.status.next',
#    right: 'git.status.next',
#    down: 'git.status.next',
    right: 'cursor.right',
    left: 'cursor.left',
    down: 'cursor.down',
    up: 'cursor.up',
    (shift, tab): 'git.status.prev',
#    left: 'git.status.prev',
#    up: 'git.status.prev',

    pagedown: 'cursor.pagedown',
    pageup: 'cursor.pageup',

    ' ': 'git.status.press',
    '\r': 'git.log.diff',
    '\n': 'git.log.diff',
    (alt, 'v'): 'git.log.prevpage',
    (alt, 't'): 'git.log.nextpage',
}


class GitLogMode(defaultmode.DefaultMode):
    # todo: rename key/command names
    SCREEN_NOWRAP = True
    DOCUMENT_MODE = False
    USE_UNDO = False

    KEY_BINDS = [
        keybind.app_keys,
        keybind.search_command_keys,
    ]

    _page_from = 0
    max_count = 100
    _commits = ()

    def init_keybind(self):
        self.register_keys(self.keybind, self.KEY_BINDS)
        self.keybind.add_keybind(log_keys)

    def init_themes(self):
        super().init_themes()
        self.themes.append(dialogmode.DialogThemes)
        self.themes.append(GitLogTheme)

    def on_str(self, wnd, s, overwrite=False):
        # does nothing
        pass

    def on_esc_pressed(self, wnd, event):
        is_available, command = self.get_command('file.close')
        command(wnd)

    def _set_log(self):
        self.document.delete(0, self.document.endpos())

        with dialogmode.FormBuilder(self.document) as f:
            f.append_text('git-log-header', 'git log')
            f.append_text('right-button', '[Pre&v 100]',
                          shortcut_style='right-button.shortcut')
            f.append_text('right-button', '[Nex&t 100]',
                          shortcut_style='right-button.shortcut')
#            f.append_text('right-button', '[&Grep]',
#                          shortcut_style='right-button.shortcut')
#            f.append_text('right-button', '[Pa&th]',
#                          shortcut_style='right-button.shortcut')
            f.append_text('git-log-header', '\n')

        default = self.get_styleid('default')
        self._commits = list(self._repo.iter_commits(max_count=self.max_count, skip=self._page_from))

        for x in self._commits:
            s = datetime.datetime.fromtimestamp(
                x.committed_date, TZ).strftime('%Y/%m/%d %H:%M:%S')
            self.document.append(s+' ', default)
            self.document.append('{:12s} '.format(x.author.name[:12]), default)
            msgs = filter(lambda l:l, (l.strip() for l in x.message.split('\n')))
            self.document.append(x.message.split('\n')[0]+'\n', default)

    def show_log(self, repo):
        self._repo = repo
        self.document.set_title('<git log>')
        self._set_log()

    @commandid('git.log.diff')
    def open_file(self, wnd):
        lineno = self.document.buf.lineno.lineno(wnd.cursor.pos)
        if lineno == 1:
            return
        commit = self._commits[lineno-2]
        d = self._repo.git.log('-1', '-c',  '--patch', commit.hexsha, pretty='format:')

        viewdiffmode.show_diff(d)

    @commandid('git.log.nextpage')
    def nextpage(self, wnd):
        self._page_from += len(self._commits)
        self._set_log()

    @commandid('git.log.prevpage')
    def prevpage(self, wnd):
        self._page_from -= self.max_count
        self._page_from = max(self._page_from, 0)
        self._set_log()

def show_git_log(d):
    repo = gitrepo.open_repo(d)
    doc = document.Document()
    mode = GitLogMode()
    doc.setmode(mode)

    mode.show_log(repo)
    kaa.app.show_doc(doc)
